import dataclasses
import datetime
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Any
from dateutil.parser import parse


class AlarmStates(Enum):
    DISARMED = "DISARMED"
    ARMED_AWAY = "ARMED_AWAY"
    ARMED_NIGHT = "ARMED_NIGHT"
    ARMED_PARTLY = "ARMED_PARTLY"
    BREACHED = "BREACHED"
    ALARM_PENDING = "ALARM_PENDING"
    ALARM_STAY_PENDING = "ALARM_STAY_PENDING"
    ARMED_NIGHT_PENDING = "ARMED_NIGHT_PENDING"
    ARMED_AWAY_PENDING = "ARMED_AWAY_PENDING"


@dataclass
class Location:
    name: str
    role: str
    user_id: str
    location_id: str
    gateway_serial: str

    def __str__(self):
        return self.name


@dataclass
class State:
    feature_name: str

    @classmethod
    def create_from_rest_response(cls, data: dict) -> "State":
        raise NotImplementedError


def extract_value_and_last_updated(data: dict) -> tuple[Any, datetime.datetime]:
    timestamp = parse(data["lastUpdated"]) if data["lastUpdated"] is not None else None
    return data["value"], timestamp


@dataclass
class BasicAlarmState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "BasicAlarmState":
        my_data = data["features"]["alarm"]["states"]
        return BasicAlarmState(
            "alarm",
            *extract_value_and_last_updated(my_data["alarm"]),
            *extract_value_and_last_updated(my_data["tamper"]),
        )

    alarm: bool
    alarm_last_updated: datetime.datetime
    tamper: bool
    tamper_last_updated: datetime.datetime


@dataclass
class SmokeAlarmState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "SmokeAlarmState":
        my_data = data["features"]["alarm"]["states"]
        return SmokeAlarmState(
            "alarm",
            *extract_value_and_last_updated(my_data["fire"]),
        )

    fire: bool
    fire_last_updated: datetime.datetime


@dataclass
class MotionSensorState(BasicAlarmState):
    sensitivity_level: Optional[float]
    sensitivity_level_last_updated: Optional[datetime.datetime]

    @classmethod
    def create_from_rest_response(cls, data: dict) -> "MotionSensorState":
        my_data = data["features"]["alarm"]["states"]
        return MotionSensorState(
            "alarm",
            *extract_value_and_last_updated(my_data["alarm"]),
            *extract_value_and_last_updated(my_data["tamper"]),
            *extract_value_and_last_updated(my_data["sensitivitylevel"]),
        )


@dataclass
class BatteryState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "BatteryState":
        my_data = data["features"]["battery"]["states"]
        return BatteryState(
            "battery",
            *extract_value_and_last_updated(my_data["low"]),
            *extract_value_and_last_updated(my_data["voltage"]),
            *extract_value_and_last_updated(my_data["defect"]) if "defect" in my_data else (None, None),
        )

    low: bool
    low_last_updated: datetime.datetime
    voltage: float
    voltage_last_updated: datetime.datetime
    defect: Optional[Any]
    defect_last_updated: datetime.datetime


@dataclass
class TemperatureState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "TemperatureState":
        my_data = data["features"]["temperature"]["states"]
        return TemperatureState(
            "temperature",
            *extract_value_and_last_updated(my_data["temperature"]),
        )

    temperature: float
    temperature_last_updated: datetime.datetime


@dataclass
class DiagnosticState(State):
    @classmethod
    def create_from_rest_response(cls, data: dict) -> "DiagnosticState":
        my_data = data["features"]["diagnostic"]["states"]
        return DiagnosticState(
            "diagnostic",
            *extract_value_and_last_updated(my_data["networklinkaddress"]),
            *extract_value_and_last_updated(my_data["networklinkstrength"]),
        )

    network_link_address: str
    network_link_address_last_updated: datetime.datetime
    network_link_strength: float
    network_link_strength_last_updated: datetime.datetime


@dataclass
class Device:
    id: str
    name: str
    serial_number: str
    location: str
    online: bool
    model_id: str
    model_name: str

    def __str__(self):
        return self.name

    @classmethod
    def create_from_rest_response(cls, device: dict) -> "Device":
        state_fields = {}
        for field in dataclasses.fields(cls):
            if issubclass(field.type, State):
                state_fields[field.name] = field.type.create_from_rest_response(device)
        return cls(
            device["id"],
            device["name"],
            device["serialNumber"],
            device["location"],
            device["online"],
            device["modelId"],
            device["modelName"],
            **state_fields,
        )

    def get_entities(self) -> list[State]:
        return [getattr(self, field.name) for field in dataclasses.fields(self) if issubclass(field.type, State)]


@dataclass
class WindowSensor(Device):
    battery: BatteryState
    diagnostic: DiagnosticState
    temperature: TemperatureState
    alarm: BasicAlarmState


@dataclass
class SmokeAlarm(Device):
    battery: BatteryState
    diagnostic: DiagnosticState
    temperature: TemperatureState
    alarm: SmokeAlarmState


@dataclass
class MotionSensorMini(Device):
    battery: BatteryState
    diagnostic: DiagnosticState
    temperature: TemperatureState
    alarm: MotionSensorState


DEVICE_MAP = {"Motion Sensor Mini": MotionSensorMini, "Smoke Alarm": SmokeAlarm, "Window Sensor": WindowSensor}


def create_device_from_rest_response(data: dict) -> Optional[Device]:
    device_class = DEVICE_MAP.get(data.get("modelName"))
    if device_class is None:
        return
    return device_class.create_from_rest_response(data)


@dataclass
class SingleLocation:
    location_id: str
    gateway_serial: str
    name: str
    alarm_state: AlarmStates
    user_role_at_location: str
    devices: list[Device]

    def __str__(self):
        return f"{self.name} with {len(self.devices)} devices"



"""
R.E.L.
Registed Event Listener
module that uses various methods to emulate event-like behavior
functions:
    read(socket, callback, *args)
    write(socket, callback, *args)
    timeout(delay, callback, *args)
    signal(sig, callback, *args)
    event(callback,arg=None,evtype=0,handle=None)
    dispatch()
    loop()
    abort()
    init()
"""
import sys, threading, time, pprint
from .registrar import set_sleep, set_turbo, SelectRegistrar, PollRegistrar, EpollRegistrar, KqueueRegistrar
from .listener import EV_PERSIST, EV_READ, EV_SIGNAL, EV_TIMEOUT, EV_WRITE
try:
    import event as pyevent
except ImportError:
    pyevent = None

def override():
    if 'event' in sys.modules and sys.modules['event'].__class__.__name__ == "fakemodule":
        return
    class fakemodule(object):
        def __init__(self, **kwargs):
            for key, val in list(kwargs.items()):
                setattr(self, key, val)
    pyevent_03_keys = [
        'EV_PERSIST',
        'EV_READ',
        'EV_SIGNAL',
        'EV_TIMEOUT',
        'EV_WRITE',
        '__author__',
        '__builtins__',
        '__copyright__',
        '__doc__',
        '__event_exc',
        '__file__',
        '__license__',
        '__name__',
        '__url__',
        '__version__',
        'abort',
        'dispatch',
        'event',
        'init',
        'loop',
        'read',
        'signal',
        'sys',
        'timeout',
        'write'
    ]
    kw = {}
    for key in pyevent_03_keys:
        kw[key] = globals().get(key, None)
    fakeevent = fakemodule(**kw)
    sys.modules['event'] = fakeevent

running = False
registrar = None
threader = None
verbose = False
supported_methods = ['epoll','poll','kqueue','select','pyevent']

mapping = {
    'select': SelectRegistrar,
    'epoll': EpollRegistrar,
    'poll': PollRegistrar,
    'kqueue': KqueueRegistrar
}

def _display(text):
    if verbose:
        print("Registered Event Listener output:", text)

def __report():
    print("=" * 60)
    print("rel status report".center(60))
    print(str(time.time()).center(60))
    print("-" * 60)
    print("events".center(60))
    pprint.pprint(registrar.events)
    print("signals".center(60))
    pprint.pprint(registrar.signals)
    print("timers".center(60))
    pprint.pprint(registrar.timers)
    print("=" * 60)
    return True

class Thread_Checker(object):
    def __init__(self, threaded):
        self.active = threaded and registrar == pyevent
        if threaded and registrar != pyevent:
            _display('GIL hack unnecessary for non-pyevent registrar. GIL hack disabled.')
        self.go()

    def go(self):
        if self.active:
            _display('Thread Checking Enabled')
            self.checker = timeout(1,self.check)
            self.sleeper = timeout(0.01, self.release)
            self.sleeper.delete()

    def stop(self):
        if self.active:
            _display('Thread Checking Disabled')
            self.checker.delete()
            self.sleeper.delete()

    def release(self, *args):
        time.sleep(.005)
        return True

    def check(self):
        if threading.activeCount() > 1:
            if not self.sleeper.pending():
                _display('Enabling GIL hack')
                self.sleeper.add(.01)
        else:
            if self.sleeper.pending():
                _display('Disabling GIL hack')
                self.sleeper.delete()
        return True

def check_init():
    if not registrar:
        initialize()

def get_registrar(method):
    if method == 'pyevent':
        if not pyevent:
            raise ImportError("could not import event")
        return pyevent
    if method in mapping:
        return mapping[method]()
    raise ImportError

def initialize(methods=supported_methods,options=()):
    """
    initialize(methods=['epoll','poll','kqueue','select','pyevent'],options=[])
    possible options:
        'verbose' - prints out certain events
        'report' - prints status of non-pyevent registrar every 5 seconds
        'strict' - ONLY try specified methods
        'threaded' - enable GIL hack -- pyevent only!
    """
    global registrar
    global threader
    global verbose
    verbose = "verbose" in options
    if "strict" not in options:
        for m in supported_methods:
            if m not in methods:
                methods.append(m)
    for method in methods:
        try:
            registrar = get_registrar(method)
            break
        except ImportError:
            _display('Could not import "%s"'%method)
    if registrar is None:
        raise ImportError("Could not import any of given methods: %s" % (methods,))
    _display('Initialized with "%s"'%method)
    threader = Thread_Checker('threaded' in options)
    if "report" in options:
        if registrar == pyevent:
            _display('Reporting disabled in pyevent. Choose epoll, poll, or select to enable reporting.')
        else:
            timeout(5,__report)
    return method

SAFE_READ = False
def safe_read():
    global SAFE_READ
    SAFE_READ = True

def read(sock,cb,*args):
    check_init()
    if SAFE_READ:
        return registrar.read(sock,cb)
    return registrar.read(sock,cb,*args)

def write(sock,cb,*args):
    check_init()
    return registrar.write(sock,cb,*args)

def timeout(delay, cb, *args):
    check_init()
    return registrar.timeout(delay,cb,*args)

def signal(sig, callback, *args):
    check_init()
    return registrar.signal(sig,callback,*args)

def dispatch():
    global running
    if not running:
        running = True
        check_init()
        registrar.dispatch()

def loop():
    check_init()
    registrar.loop()

def report():
    check_init()
    return registrar.report()

def abort():
    global running
    running = False
    check_init()
    registrar.abort()

def abort_branch():
    check_init()
    registrar.abort_branch()

def init():
    global running
    running = False
    check_init()
    threader.stop()
    registrar.init()
    threader.go()

def event(callback,arg=None,evtype=0,handle=None):
    check_init()
    return registrar.event(callback,arg,evtype,handle)

def _thread_wrapper(callback):
    from .errors import AbortBranch
    try:
        callback()
    except AbortBranch as e:
        pass # we don't care at this point

def thread(callback):
    threading.Thread(target=_thread_wrapper, args=(callback,)).start()

def tick():
    check_init()
    return registrar.tick


import datetime
from unittest import TestCase

from dateutil.tz import tzutc

from homelypy.devices import create_device_from_rest_response, WindowSensor, SmokeAlarm, MotionSensorMini


class TestDeviceCreation(TestCase):
    def test_create_window_sensor(self):
        rest_response = {
            "features": {
                "alarm": {
                    "states": {
                        "alarm": {"lastUpdated": "2022-12-31T16:34:31.189Z", "value": False},
                        "tamper": {"lastUpdated": "2022-06-10T15:43:20.402Z", "value": False},
                    }
                },
                "battery": {
                    "states": {
                        "defect": {"lastUpdated": None, "value": None},
                        "low": {"lastUpdated": "2022-06-10T15:29:20.956Z", "value": False},
                        "voltage": {"lastUpdated": "2022-12-09T12:33:11.390Z", "value": 2.9},
                    }
                },
                "diagnostic": {
                    "states": {
                        "networklinkaddress": {"lastUpdated": "2022-11-19T22:00:31.223Z", "value": "0015BC0041001B88"},
                        "networklinkstrength": {"lastUpdated": "2022-12-31T16:07:13.769Z", "value": 92},
                    }
                },
                "temperature": {"states": {"temperature": {"lastUpdated": "2022-12-31T16:26:12.692Z", "value": 16}}},
            },
            "id": "f6210e83-a41c-49c6-a24a-57733ba8ea44",
            "location": "Floor 0 - Entrance",
            "modelId": "87fa1ae0-824f-4d42-be7a-cc5b6c7b1e35",
            "modelName": "Window Sensor",
            "name": "Window Sensor",
            "online": True,
            "serialNumber": "0015BC001E014469",
        }
        device: WindowSensor = create_device_from_rest_response(rest_response)
        self.assertTrue(isinstance(device, WindowSensor))
        self.assertEqual("Window Sensor", device.name)
        self.assertEqual(16, device.temperature.temperature)
        self.assertEqual(
            datetime.datetime(2022, 12, 31, 16, 26, 12, 692000, tzinfo=tzutc()),
            device.temperature.temperature_last_updated,
        )

    def test_create_smoke_alarm(self):
        rest_response = {
            "features": {
                "alarm": {"states": {"fire": {"lastUpdated": "2022-12-15T10:41:00.825Z", "value": False}}},
                "battery": {
                    "states": {
                        "low": {"lastUpdated": "2022-06-10T15:30:20.675Z", "value": False},
                        "voltage": {"lastUpdated": "2022-12-24T00:44:01.043Z", "value": 3},
                    }
                },
                "diagnostic": {
                    "states": {
                        "networklinkaddress": {"lastUpdated": "2022-12-23T21:40:30.214Z", "value": "0015BC002C101A48"},
                        "networklinkstrength": {"lastUpdated": "2022-12-31T16:17:42.676Z", "value": 47},
                    }
                },
                "temperature": {"states": {"temperature": {"lastUpdated": "2022-12-31T16:30:06.306Z", "value": 17.6}}},
            },
            "id": "c90f6b7e-c451-498e-a1c6-ba3b46150ce5",
            "location": "Floor 0 - Living room",
            "modelId": "ffe30099-92c5-4471-879f-41f412d423ab",
            "modelName": "Smoke Alarm",
            "name": "Smoke Alarm",
            "online": True,
            "serialNumber": "0015BC003100CE07",
        }
        device: SmokeAlarm = create_device_from_rest_response(rest_response)
        self.assertTrue(isinstance(device, SmokeAlarm))
        self.assertFalse(device.alarm.fire)

    def test_create_motion_sensor_mini(self):
        rest_response = {
            "features": {
                "alarm": {
                    "states": {
                        "alarm": {"lastUpdated": "2022-12-31T16:28:43.861Z", "value": True},
                        "sensitivitylevel": {"lastUpdated": None, "value": None},
                        "tamper": {"lastUpdated": "2022-06-10T16:17:58.161Z", "value": False},
                    }
                },
                "battery": {
                    "states": {
                        "defect": {"lastUpdated": "2022-06-10T16:15:05.770Z", "value": False},
                        "low": {"lastUpdated": "2022-06-10T16:15:05.741Z", "value": False},
                        "voltage": {"lastUpdated": "2022-12-22T06:50:30.112Z", "value": 2.9},
                    }
                },
                "diagnostic": {
                    "states": {
                        "networklinkaddress": {"lastUpdated": "2022-12-22T05:50:26.083Z", "value": "0015BC002C101A48"},
                        "networklinkstrength": {"lastUpdated": "2022-12-31T16:27:48.088Z", "value": 89},
                    }
                },
                "temperature": {"states": {"temperature": {"lastUpdated": "2022-12-31T16:27:03.967Z", "value": 19.4}}},
            },
            "id": "28e0b340-26a6-475c-a419-a5f31bc8f479",
            "location": "Floor 1 - Hallway",
            "modelId": "e806ca73-4be0-4bd2-98cb-71f273b09812",
            "modelName": "Motion Sensor Mini",
            "name": "Motion Sensor Mini",
            "online": True,
            "serialNumber": "0015BC001A012223",
        }
        device: MotionSensorMini = create_device_from_rest_response(rest_response)
        self.assertTrue(isinstance(device, MotionSensorMini))
        self.assertEqual(89, device.diagnostic.network_link_strength)



import dataclasses
import logging
import time
from typing import Callable, Dict, List

import requests
import websocket

from homelypy.devices import Location, SingleLocation, create_device_from_rest_response

WEB_SOCKET_URL = "ws://sdk.iotiliti.cloud"

SDK_URL = "https://sdk.iotiliti.cloud"
AUTHENTICATION_ENDPOINT = "/homely/oauth/token"
LOCATIONS_ENDPOINT = "/homely/locations"
SINGLE_LOCATION_ENDPOINT = "/homely/home"
REFRESH_TOKEN_ENDPOINT = "/homely/oauth/refresh-token"

logger = logging.getLogger(__name__)


class ConnectionFailedException(Exception):
    pass


class AuthenticationFailedException(Exception):
    pass


class Homely:
    def __init__(self, username: str, password: str):
        self.refresh_expires_in = 0
        self.expires_in = 0
        self.authentication_time = 0
        self.access_token = None
        self.refresh_token = None
        self.username = username
        self.password = password

    @staticmethod
    def url(endpoint: str) -> str:
        return SDK_URL + endpoint

    def authenticate(self):
        response = requests.post(
            self.url(AUTHENTICATION_ENDPOINT),
            data={"username": self.username, "password": self.password},
        )
        if response.status_code == 401:
            raise AuthenticationFailedException(response.text)
        if response.status_code != 201:
            raise ConnectionFailedException(response.text)
        data = response.json()
        self.store_authentication_information(data)

    def store_authentication_information(self, data: Dict):
        self.access_token = data["access_token"]
        self.authentication_time = time.time()
        self.expires_in = data["expires_in"]
        self.refresh_token = data["refresh_token"]
        self.refresh_expires_in = data["refresh_expires_in"]

    def reauthenticate(self):
        response = requests.post(
            self.url(REFRESH_TOKEN_ENDPOINT),
            data={"refresh_token": self.refresh_token},
        )
        if response.status_code != 201:
            self.refresh_expires_in = -1
            self.expires_in = -1
            raise ConnectionFailedException(response.text)
        data = response.json()
        self.store_authentication_information(data)

    def authenticate_if_required(self):
        if time.time() - self.authentication_time > self.refresh_expires_in - 2:
            self.authenticate()
        elif time.time() - self.authentication_time > self.expires_in - 2:
            self.reauthenticate()

    @property
    def authorisation_header(self) -> Dict:
        self.authenticate_if_required()
        return {"Authorization": f"Bearer {self.access_token}"}

    def get_locations(self) -> List[Location]:
        response = requests.get(self.url(LOCATIONS_ENDPOINT), headers=self.authorisation_header)
        if response.status_code != 200:
            raise ConnectionFailedException(response.text)
        return [
            Location(
                location_data["name"],
                location_data["role"],
                location_data["userId"],
                location_data["locationId"],
                location_data["gatewayserial"],
            )
            for location_data in response.json()
        ]

    def get_location(self, location_id) -> SingleLocation:
        response = requests.get(
            self.url(SINGLE_LOCATION_ENDPOINT) + f"/{location_id}",
            headers=self.authorisation_header,
        )
        if response.status_code != 200:
            raise ConnectionFailedException(response.text)
        data = response.json()
        devices = []
        for device in data["devices"]:
            devices.append(create_device_from_rest_response(device))
        return SingleLocation(
            data["locationId"],
            data["gatewayserial"],
            data["name"],
            data["alarmState"],
            data["userRoleAtLocation"],
            devices,
        )

    def get_web_socket(self, location_id: str, on_message: Callable) -> websocket.WebSocketApp:
        self.authenticate_if_required()
        websocket.enableTrace(True)
        url = f"{WEB_SOCKET_URL}?locationId={location_id}"
        logger.debug(f"Connecting to web socket {url}")
        ws = websocket.WebSocketApp(
            url,
            header={**self.authorisation_header, "locationId": location_id},
            on_open=on_open,
            on_message=on_message,
            on_error=on_error,
            on_close=on_close,
        )
        return ws


def on_error(ws, error):
    logger.debug(f"Websocket error: {error}")


def on_close(ws, close_status_code, close_msg):
    logger.debug(f"Websocket closed")


def on_open(ws):
    logger.debug(f"Websocket connected")


if __name__ == "__main__":
    import rel

    logging.basicConfig(level=logging.DEBUG)
    homely = Homely("***", "***")
    locations = homely.get_locations()
    for location in locations:
        logger.debug(f"Received location '{location}'")
    location = homely.get_location(locations[0].location_id)
    logger.debug(f"Received single location '{location}'")
    for device in location.devices:
        logger.debug(f"Device: {device} is of type {device.__class__}")

    ws = homely.get_web_socket(location.location_id, lambda data: logger.debug(f"Received data: {data}"))
    ws.run_forever(
        dispatcher=rel, reconnect=5
    )  # Set dispatcher to automatic reconnection, 5 second reconnect delay if connection closed unexpectedly
    rel.signal(2, rel.abort)  # Keyboard Interrupt
    rel.dispatch()
